@@ -12,15 +12,9 @@
 #endif
 
 #ifdef FEAT_GUI_GTK
-# ifdef VMS // undef MIN and MAX because Intrinsic.h redefines them anyway
-#  ifdef MAX
-#   undef MAX
-#  endif
-#  ifdef MIN
-#   undef MIN
-#  endif
+# ifdef VMS
 #  include "gui_gtk_vms.h"
-# endif // VMS
+# endif
 # include <X11/Intrinsic.h>
 # pragma GCC diagnostic push
 # pragma GCC diagnostic ignored "-Wstrict-prototypes"
@@ -182,7 +176,7 @@ typedef struct GuiScrollbar
 				// to reduce the count.
 #endif
 
-#if FEAT_GUI_HAIKU
+#ifdef FEAT_GUI_HAIKU
     VimScrollBar *id;		// Pointer to real scroll bar
 #endif
 #ifdef FEAT_GUI_PHOTON
@@ -240,7 +234,7 @@ typedef struct Gui
     int		in_focus;	    // Vim has input focus
     int		in_use;		    // Is the GUI being used?
     int		starting;	    // GUI will start in a little while
-    int		shell_created;	    // Has the shell been created yet?
+    bool	shell_created;	    // Has the shell been created yet?
     int		dying;		    // Is vim dying? Then output to terminal
     int		dofork;		    // Use fork() when GUI is starting
 #ifdef GUI_MAY_SPAWN
@@ -265,6 +259,7 @@ typedef struct Gui
     int		scrollbar_height;   // Height of horizontal scrollbar
     int		left_sbar_x;	    // Calculated x coord for left scrollbar
     int		right_sbar_x;	    // Calculated x coord for right scrollbar
+    int         force_redraw;       // Force a redraw even e.g. not resized
 
 #ifdef FEAT_MENU
 # ifndef FEAT_GUI_GTK
@@ -394,10 +389,15 @@ typedef struct Gui
     char_u	*browse_fname;	    // file name from filedlg
 
     guint32	event_time;
+# ifdef GDK_WINDOWING_WAYLAND
+    _Bool	is_wayland;           // active gdk backend in gtk is wayland
+# endif
+#endif	// FEAT_GUI_GTK
 
+#if defined(FEAT_GUI_GTK) || defined(FEAT_GUI_MSWIN)
     char_u ligatures_map[256];	    // ascii map for characters 0-255, value is
 				    // 1 if in 'guiligatures'
-#endif	// FEAT_GUI_GTK
+#endif
 
 #if defined(FEAT_GUI_TABLINE) \
 	&& (defined(FEAT_GUI_MSWIN) || defined(FEAT_GUI_MOTIF) \
@@ -425,6 +425,11 @@ typedef struct Gui
     guicolor_T	currFgColor;	    // Current foreground text color
     guicolor_T	currBgColor;	    // Current background text color
     guicolor_T	currSpColor;	    // Current special text color
+
+    guicolor_T	title_bg_pixel;	    // window title bar color
+    guicolor_T	title_fg_pixel;	    // window title bar's text color
+    guicolor_T	titlenc_bg_pixel;   // window title bar color not current
+    guicolor_T	titlenc_fg_pixel;   // window title bar's text color not current
 #endif
 
 #ifdef FEAT_GUI_HAIKU
