@@ -14,7 +14,7 @@
 #include "vim.h"
 #include "version.h"
 
-#if defined(FEAT_PRINTER) || defined(PROTO)
+#if defined(FEAT_PRINTER)
 /*
  * To implement printing on a platform, the following functions must be
  * defined:
@@ -144,18 +144,18 @@ static colnr_T hardcopy_line(prt_settings_T *psettings, int page_line, prt_pos_T
  * Returns an error message or NULL;
  */
     char *
-parse_printoptions(void)
+parse_printoptions(optset_T *args UNUSED)
 {
     return parse_list_options(p_popt, printer_opts, OPT_PRINT_NUM_OPTIONS);
 }
 
-#if defined(FEAT_POSTSCRIPT) || defined(PROTO)
+#if defined(FEAT_POSTSCRIPT)
 /*
  * Parse 'printmbfont' and set the flags in "mbfont_opts".
  * Returns an error message or NULL;
  */
     char *
-parse_printmbfont(void)
+parse_printmbfont(optset_T *args UNUSED)
 {
     return parse_list_options(p_pmfn, mbfont_opts, OPT_MBFONT_NUM_OPTIONS);
 }
@@ -293,6 +293,7 @@ prt_get_attr(
     pattr->italic = (highlight_has_attr(hl_id, HL_ITALIC, modec) != NULL);
     pattr->underline = (highlight_has_attr(hl_id, HL_UNDERLINE, modec) != NULL);
     pattr->undercurl = (highlight_has_attr(hl_id, HL_UNDERCURL, modec) != NULL);
+    // TODO: HL_UNDERDOUBLE, HL_UNDERDOTTED, HL_UNDERDASHED
 
 # if defined(FEAT_GUI) || defined(FEAT_TERMGUICOLORS)
     if (USE_24BIT)
@@ -333,35 +334,35 @@ prt_get_attr(
     static void
 prt_set_fg(long_u fg)
 {
-    if (fg != curr_fg)
-    {
-	curr_fg = fg;
-	mch_print_set_fg(fg);
-    }
+    if (fg == curr_fg)
+	return;
+
+    curr_fg = fg;
+    mch_print_set_fg(fg);
 }
 
     static void
 prt_set_bg(long_u bg)
 {
-    if (bg != curr_bg)
-    {
-	curr_bg = bg;
-	mch_print_set_bg(bg);
-    }
+    if (bg == curr_bg)
+	return;
+
+    curr_bg = bg;
+    mch_print_set_bg(bg);
 }
 
     static void
 prt_set_font(int bold, int italic, int underline)
 {
-    if (curr_bold != bold
-	    || curr_italic != italic
-	    || curr_underline != underline)
-    {
-	curr_underline = underline;
-	curr_italic = italic;
-	curr_bold = bold;
-	mch_print_set_font(bold, italic, underline);
-    }
+    if (curr_bold == bold
+	    && curr_italic == italic
+	    && curr_underline == underline)
+	return;
+
+    curr_underline = underline;
+    curr_italic = italic;
+    curr_bold = bold;
+    mch_print_set_font(bold, italic, underline);
 }
 
 /*
@@ -433,13 +434,15 @@ prt_get_unit(int idx)
     int		i;
     static char *(units[4]) = PRT_UNIT_NAMES;
 
-    if (printer_opts[idx].present)
-	for (i = 0; i < 4; ++i)
-	    if (STRNICMP(printer_opts[idx].string, units[i], 2) == 0)
-	    {
-		u = i;
-		break;
-	    }
+    if (!printer_opts[idx].present)
+	return PRT_UNIT_NONE;
+
+    for (i = 0; i < 4; ++i)
+	if (STRNICMP(printer_opts[idx].string, units[i], 2) == 0)
+	{
+	    u = i;
+	    break;
+	}
     return u;
 }
 
@@ -470,7 +473,6 @@ prt_header(
     if (*p_header != NUL)
     {
 	linenr_T	tmp_lnum, tmp_topline, tmp_botline;
-	int		use_sandbox = FALSE;
 
 	/*
 	 * Need to (temporarily) set current line number and first/last line
@@ -486,12 +488,8 @@ prt_header(
 	curwin->w_botline = lnum + 63;
 	printer_page_num = pagenum;
 
-# ifdef FEAT_EVAL
-	use_sandbox = was_set_insecurely((char_u *)"printheader", 0);
-# endif
-	build_stl_str_hl(curwin, tbuf, (size_t)(width + IOSIZE),
-						  p_header, use_sandbox,
-						  ' ', width, NULL, NULL);
+	build_stl_str_hl(curwin, tbuf, (size_t)(width + IOSIZE), p_header,
+			   (char_u *)"printheader", 0, ' ', width, NULL, NULL);
 
 	// Reset line numbers
 	curwin->w_cursor.lnum = tmp_lnum;
@@ -561,7 +559,7 @@ ex_hardcopy(exarg_T *eap)
     CLEAR_FIELD(settings);
     settings.has_color = TRUE;
 
-# ifdef FEAT_POSTSCRIPT
+#ifdef FEAT_POSTSCRIPT
     if (*eap->arg == '>')
     {
 	char	*errormsg = NULL;
@@ -577,7 +575,7 @@ ex_hardcopy(exarg_T *eap)
     }
     else if (*eap->arg != NUL)
 	settings.arguments = eap->arg;
-# endif
+#endif
 
     /*
      * Initialise for printing.  Ask the user for settings, unless forceit is
@@ -941,7 +939,7 @@ hardcopy_line(
     return col;
 }
 
-# if defined(FEAT_POSTSCRIPT) || defined(PROTO)
+#if defined(FEAT_POSTSCRIPT)
 
 /*
  * PS printer stuff.
@@ -972,9 +970,9 @@ hardcopy_line(
  * http://www.adobe.com
  */
 
-#define PRT_PS_DEFAULT_DPI	    (72)    // Default user space resolution
-#define PRT_PS_DEFAULT_FONTSIZE     (10)
-#define PRT_PS_DEFAULT_BUFFER_SIZE  (80)
+# define PRT_PS_DEFAULT_DPI	    (72)    // Default user space resolution
+# define PRT_PS_DEFAULT_FONTSIZE     (10)
+# define PRT_PS_DEFAULT_BUFFER_SIZE  (80)
 
 struct prt_mediasize_S
 {
@@ -983,7 +981,7 @@ struct prt_mediasize_S
     float	height;
 };
 
-#define PRT_MEDIASIZE_LEN  ARRAY_LENGTH(prt_mediasize)
+# define PRT_MEDIASIZE_LEN  ARRAY_LENGTH(prt_mediasize)
 
 static struct prt_mediasize_S prt_mediasize[] =
 {
@@ -1014,10 +1012,10 @@ struct prt_ps_font_S
     char	*(ps_fontname[4]);
 };
 
-#define PRT_PS_FONT_ROMAN	(0)
-#define PRT_PS_FONT_BOLD	(1)
-#define PRT_PS_FONT_OBLIQUE	(2)
-#define PRT_PS_FONT_BOLDOBLIQUE (3)
+# define PRT_PS_FONT_ROMAN	(0)
+# define PRT_PS_FONT_BOLD	(1)
+# define PRT_PS_FONT_OBLIQUE	(2)
+# define PRT_PS_FONT_BOLDOBLIQUE (3)
 
 // Standard font metrics for Courier family
 static struct prt_ps_font_S prt_ps_courier_font =
@@ -1057,14 +1055,14 @@ struct prt_ps_charset_S
 };
 
 
-#define CS_JIS_C_1978   (0x01)
-#define CS_JIS_X_1983   (0x02)
-#define CS_JIS_X_1990   (0x04)
-#define CS_NEC		(0x08)
-#define CS_MSWINDOWS	(0x10)
-#define CS_CP932	(0x20)
-#define CS_KANJITALK6	(0x40)
-#define CS_KANJITALK7   (0x80)
+# define CS_JIS_C_1978   (0x01)
+# define CS_JIS_X_1983   (0x02)
+# define CS_JIS_X_1990   (0x04)
+# define CS_NEC		(0x08)
+# define CS_MSWINDOWS	(0x10)
+# define CS_CP932	(0x20)
+# define CS_KANJITALK6	(0x40)
+# define CS_KANJITALK7   (0x80)
 
 // Japanese encodings and charsets
 static struct prt_ps_encoding_S j_encodings[] =
@@ -1090,13 +1088,13 @@ static struct prt_ps_charset_S j_charsets[] =
     {"KANJITALK7",  "90pv",     CS_KANJITALK7}
 };
 
-#define CS_GB_2312_80       (0x01)
-#define CS_GBT_12345_90     (0x02)
-#define CS_GBK2K	    (0x04)
-#define CS_SC_MAC	    (0x08)
-#define CS_GBT_90_MAC	    (0x10)
-#define CS_GBK		    (0x20)
-#define CS_SC_ISO10646      (0x40)
+# define CS_GB_2312_80       (0x01)
+# define CS_GBT_12345_90     (0x02)
+# define CS_GBK2K	    (0x04)
+# define CS_SC_MAC	    (0x08)
+# define CS_GBT_90_MAC	    (0x10)
+# define CS_GBK		    (0x20)
+# define CS_SC_ISO10646      (0x40)
 
 // Simplified Chinese encodings and charsets
 static struct prt_ps_encoding_S sc_encodings[] =
@@ -1120,19 +1118,19 @@ static struct prt_ps_charset_S sc_charsets[] =
     {"ISO10646",    "UniGB",    CS_SC_ISO10646}
 };
 
-#define CS_CNS_PLANE_1      (0x01)
-#define CS_CNS_PLANE_2      (0x02)
-#define CS_CNS_PLANE_1_2    (0x04)
-#define CS_B5		    (0x08)
-#define CS_ETEN		    (0x10)
-#define CS_HK_GCCS	    (0x20)
-#define CS_HK_SCS	    (0x40)
-#define CS_HK_SCS_ETEN	    (0x80)
-#define CS_MTHKL	    (0x100)
-#define CS_MTHKS	    (0x200)
-#define CS_DLHKL	    (0x400)
-#define CS_DLHKS	    (0x800)
-#define CS_TC_ISO10646	    (0x1000)
+# define CS_CNS_PLANE_1      (0x01)
+# define CS_CNS_PLANE_2      (0x02)
+# define CS_CNS_PLANE_1_2    (0x04)
+# define CS_B5		    (0x08)
+# define CS_ETEN		    (0x10)
+# define CS_HK_GCCS	    (0x20)
+# define CS_HK_SCS	    (0x40)
+# define CS_HK_SCS_ETEN	    (0x80)
+# define CS_MTHKL	    (0x100)
+# define CS_MTHKS	    (0x200)
+# define CS_DLHKL	    (0x400)
+# define CS_DLHKS	    (0x800)
+# define CS_TC_ISO10646	    (0x1000)
 
 // Traditional Chinese encodings and charsets
 static struct prt_ps_encoding_S tc_encodings[] =
@@ -1166,10 +1164,10 @@ static struct prt_ps_charset_S tc_charsets[] =
     {"ISO10646",    "UniCNS",   CS_TC_ISO10646}
 };
 
-#define CS_KR_X_1992	    (0x01)
-#define CS_KR_MAC	    (0x02)
-#define CS_KR_X_1992_MS     (0x04)
-#define CS_KR_ISO10646      (0x08)
+# define CS_KR_X_1992	    (0x01)
+# define CS_KR_MAC	    (0x02)
+# define CS_KR_X_1992_MS     (0x04)
+# define CS_KR_ISO10646      (0x08)
 
 // Korean encodings and charsets
 static struct prt_ps_encoding_S k_encodings[] =
@@ -1251,9 +1249,9 @@ struct prt_ps_resource_S
 };
 
 // Types of PS resource file currently used
-#define PRT_RESOURCE_TYPE_PROCSET   (0)
-#define PRT_RESOURCE_TYPE_ENCODING  (1)
-#define PRT_RESOURCE_TYPE_CMAP      (2)
+# define PRT_RESOURCE_TYPE_PROCSET   (0)
+# define PRT_RESOURCE_TYPE_ENCODING  (1)
+# define PRT_RESOURCE_TYPE_CMAP      (2)
 
 // The PS prolog file version number has to match - if the prolog file is
 // updated, increment the number in the file and here.  Version checking was
@@ -1264,8 +1262,8 @@ struct prt_ps_resource_S
 // VIM      Prolog  CIDProlog
 // 6.2      1.3
 // 7.0      1.4	    1.0
-#define PRT_PROLOG_VERSION  ((char_u *)"1.4")
-#define PRT_CID_PROLOG_VERSION  ((char_u *)"1.0")
+# define PRT_PROLOG_VERSION  ((char_u *)"1.4")
+# define PRT_CID_PROLOG_VERSION  ((char_u *)"1.0")
 
 // String versions of PS resource types - indexed by constants above so don't
 // re-order!
@@ -1277,23 +1275,23 @@ static char *prt_resource_types[] =
 };
 
 // Strings to look for in a PS resource file
-#define PRT_RESOURCE_HEADER	    "%!PS-Adobe-"
-#define PRT_RESOURCE_RESOURCE	    "Resource-"
-#define PRT_RESOURCE_PROCSET	    "ProcSet"
-#define PRT_RESOURCE_ENCODING	    "Encoding"
-#define PRT_RESOURCE_CMAP	    "CMap"
+# define PRT_RESOURCE_HEADER	    "%!PS-Adobe-"
+# define PRT_RESOURCE_RESOURCE	    "Resource-"
+# define PRT_RESOURCE_PROCSET	    "ProcSet"
+# define PRT_RESOURCE_ENCODING	    "Encoding"
+# define PRT_RESOURCE_CMAP	    "CMap"
 
 
 // Data for table based DSC comment recognition, easy to extend if VIM needs to
 // read more comments.
-#define PRT_DSC_MISC_TYPE	    (-1)
-#define PRT_DSC_TITLE_TYPE	    (1)
-#define PRT_DSC_VERSION_TYPE	    (2)
-#define PRT_DSC_ENDCOMMENTS_TYPE    (3)
+# define PRT_DSC_MISC_TYPE	    (-1)
+# define PRT_DSC_TITLE_TYPE	    (1)
+# define PRT_DSC_VERSION_TYPE	    (2)
+# define PRT_DSC_ENDCOMMENTS_TYPE    (3)
 
-#define PRT_DSC_TITLE		    "%%Title:"
-#define PRT_DSC_VERSION		    "%%Version:"
-#define PRT_DSC_ENDCOMMENTS	    "%%EndComments:"
+# define PRT_DSC_TITLE		    "%%Title:"
+# define PRT_DSC_VERSION		    "%%Version:"
+# define PRT_DSC_ENDCOMMENTS	    "%%EndComments:"
 
 struct prt_dsc_comment_S
 {
@@ -1310,7 +1308,7 @@ struct prt_dsc_line_S
 };
 
 
-#define SIZEOF_CSTR(s)      (sizeof(s) - 1)
+# define SIZEOF_CSTR(s)      (sizeof(s) - 1)
 static struct prt_dsc_comment_S prt_dsc_table[] =
 {
     {PRT_DSC_TITLE,       SIZEOF_CSTR(PRT_DSC_TITLE),     PRT_DSC_TITLE_TYPE},
@@ -1419,9 +1417,6 @@ prt_write_file(char_u *buffer)
     static void
 prt_write_file_len(char_u *buffer, int bytes)
 {
-#ifdef EBCDIC
-    ebcdic2ascii(buffer, bytes);
-#endif
     prt_write_file_raw_len(buffer, bytes);
 }
 
@@ -1576,82 +1571,80 @@ prt_def_var(char *name, double value, int prec)
 }
 
 // Convert size from font space to user space at current font scale
-#define PRT_PS_FONT_TO_USER(scale, size)    ((size) * ((scale)/1000.0))
+# define PRT_PS_FONT_TO_USER(scale, size)    ((size) * ((scale)/1000.0))
 
     static void
 prt_flush_buffer(void)
 {
-    if (prt_ps_buffer.ga_len > 0)
+    if (prt_ps_buffer.ga_len <= 0)
+	return;
+
+    // Any background color must be drawn first
+    if (prt_do_bgcol && (prt_new_bgcol != PRCOLOR_WHITE))
     {
-	// Any background color must be drawn first
-	if (prt_do_bgcol && (prt_new_bgcol != PRCOLOR_WHITE))
-	{
-	    int     r, g, b;
+	int     r, g, b;
 
-	    if (prt_do_moveto)
-	    {
-		prt_write_real(prt_pos_x_moveto, 2);
-		prt_write_real(prt_pos_y_moveto, 2);
-		prt_write_string("m\n");
-		prt_do_moveto = FALSE;
-	    }
-
-	    // Size of rect of background color on which text is printed
-	    prt_write_real(prt_text_run, 2);
-	    prt_write_real(prt_line_height, 2);
-
-	    // Lastly add the color of the background
-	    r = ((unsigned)prt_new_bgcol & 0xff0000) >> 16;
-	    g = ((unsigned)prt_new_bgcol & 0xff00) >> 8;
-	    b = prt_new_bgcol & 0xff;
-	    prt_write_real(r / 255.0, 3);
-	    prt_write_real(g / 255.0, 3);
-	    prt_write_real(b / 255.0, 3);
-	    prt_write_string("bg\n");
-	}
-	// Draw underlines before the text as it makes it slightly easier to
-	// find the starting point.
-	if (prt_do_underline)
-	{
-	    if (prt_do_moveto)
-	    {
-		prt_write_real(prt_pos_x_moveto, 2);
-		prt_write_real(prt_pos_y_moveto, 2);
-		prt_write_string("m\n");
-		prt_do_moveto = FALSE;
-	    }
-
-	    // Underline length of text run
-	    prt_write_real(prt_text_run, 2);
-	    prt_write_string("ul\n");
-	}
-	// Draw the text
-	// Note: we write text out raw - EBCDIC conversion is handled in the
-	// PostScript world via the font encoding vector.
-	if (prt_out_mbyte)
-	    prt_write_string("<");
-	else
-	    prt_write_string("(");
-	prt_write_file_raw_len(prt_ps_buffer.ga_data, prt_ps_buffer.ga_len);
-	if (prt_out_mbyte)
-	    prt_write_string(">");
-	else
-	    prt_write_string(")");
-	// Add a moveto if need be and use the appropriate show procedure
 	if (prt_do_moveto)
 	{
 	    prt_write_real(prt_pos_x_moveto, 2);
 	    prt_write_real(prt_pos_y_moveto, 2);
-	    // moveto and a show
-	    prt_write_string("ms\n");
+	    prt_write_string("m\n");
 	    prt_do_moveto = FALSE;
 	}
-	else // Simple show
-	    prt_write_string("s\n");
 
-	ga_clear(&prt_ps_buffer);
-	ga_init2(&prt_ps_buffer, sizeof(char), prt_bufsiz);
+	// Size of rect of background color on which text is printed
+	prt_write_real(prt_text_run, 2);
+	prt_write_real(prt_line_height, 2);
+
+	// Lastly add the color of the background
+	r = ((unsigned)prt_new_bgcol & 0xff0000) >> 16;
+	g = ((unsigned)prt_new_bgcol & 0xff00) >> 8;
+	b = prt_new_bgcol & 0xff;
+	prt_write_real(r / 255.0, 3);
+	prt_write_real(g / 255.0, 3);
+	prt_write_real(b / 255.0, 3);
+	prt_write_string("bg\n");
     }
+    // Draw underlines before the text as it makes it slightly easier to
+    // find the starting point.
+    if (prt_do_underline)
+    {
+	if (prt_do_moveto)
+	{
+	    prt_write_real(prt_pos_x_moveto, 2);
+	    prt_write_real(prt_pos_y_moveto, 2);
+	    prt_write_string("m\n");
+	    prt_do_moveto = FALSE;
+	}
+
+	// Underline length of text run
+	prt_write_real(prt_text_run, 2);
+	prt_write_string("ul\n");
+    }
+    // Draw the text
+    if (prt_out_mbyte)
+	prt_write_string("<");
+    else
+	prt_write_string("(");
+    prt_write_file_raw_len(prt_ps_buffer.ga_data, prt_ps_buffer.ga_len);
+    if (prt_out_mbyte)
+	prt_write_string(">");
+    else
+	prt_write_string(")");
+    // Add a moveto if need be and use the appropriate show procedure
+    if (prt_do_moveto)
+    {
+	prt_write_real(prt_pos_x_moveto, 2);
+	prt_write_real(prt_pos_y_moveto, 2);
+	// moveto and a show
+	prt_write_string("ms\n");
+	prt_do_moveto = FALSE;
+    }
+    else // Simple show
+	prt_write_string("s\n");
+
+    ga_clear(&prt_ps_buffer);
+    ga_init2(&prt_ps_buffer, sizeof(char), prt_bufsiz);
 }
 
 
@@ -1691,12 +1684,12 @@ prt_find_resource(char *name, struct prt_ps_resource_S *resource)
 }
 
 // PS CR and LF characters have platform independent values
-#define PSLF  (0x0a)
-#define PSCR  (0x0d)
+# define PSLF  (0x0a)
+# define PSCR  (0x0d)
 
 // Static buffer to read initial comments in a resource file, some can have a
 // couple of KB of comments!
-#define PRT_FILE_BUFFER_LEN (2048)
+# define PRT_FILE_BUFFER_LEN (2048)
 struct prt_resfile_buffer_S
 {
     char_u  buffer[PRT_FILE_BUFFER_LEN];
@@ -1752,7 +1745,7 @@ prt_resfile_skip_nonws(int offset)
     idx = prt_resfile.line_start + offset;
     while (idx < prt_resfile.line_end)
     {
-	if (isspace(prt_resfile.buffer[idx]))
+	if (SAFE_isspace(prt_resfile.buffer[idx]))
 	    return idx - prt_resfile.line_start;
 	idx++;
     }
@@ -1767,7 +1760,7 @@ prt_resfile_skip_ws(int offset)
     idx = prt_resfile.line_start + offset;
     while (idx < prt_resfile.line_end)
     {
-	if (!isspace(prt_resfile.buffer[idx]))
+	if (!SAFE_isspace(prt_resfile.buffer[idx]))
 	    return idx - prt_resfile.line_start;
 	idx++;
     }
@@ -1864,7 +1857,7 @@ prt_open_resource(struct prt_ps_resource_S *resource)
     offset = 0;
 
     if (prt_resfile_strncmp(offset, PRT_RESOURCE_HEADER,
-				       (int)STRLEN(PRT_RESOURCE_HEADER)) != 0)
+			       (int)STRLEN_LITERAL(PRT_RESOURCE_HEADER)) != 0)
     {
 	semsg(_(e_file_str_is_not_postscript_resource_file),
 							   resource->filename);
@@ -1872,7 +1865,7 @@ prt_open_resource(struct prt_ps_resource_S *resource)
     }
 
     // Skip over any version numbers and following ws
-    offset += (int)STRLEN(PRT_RESOURCE_HEADER);
+    offset += (int)STRLEN_LITERAL(PRT_RESOURCE_HEADER);
     offset = prt_resfile_skip_nonws(offset);
     if (offset == -1)
 	return FALSE;
@@ -1881,23 +1874,23 @@ prt_open_resource(struct prt_ps_resource_S *resource)
 	return FALSE;
 
     if (prt_resfile_strncmp(offset, PRT_RESOURCE_RESOURCE,
-				     (int)STRLEN(PRT_RESOURCE_RESOURCE)) != 0)
+			     (int)STRLEN_LITERAL(PRT_RESOURCE_RESOURCE)) != 0)
     {
 	semsg(_(e_file_str_is_not_supported_postscript_resource_file),
 							   resource->filename);
 	return FALSE;
     }
-    offset += (int)STRLEN(PRT_RESOURCE_RESOURCE);
+    offset += (int)STRLEN_LITERAL(PRT_RESOURCE_RESOURCE);
 
     // Decide type of resource in the file
     if (prt_resfile_strncmp(offset, PRT_RESOURCE_PROCSET,
-				      (int)STRLEN(PRT_RESOURCE_PROCSET)) == 0)
+			      (int)STRLEN_LITERAL(PRT_RESOURCE_PROCSET)) == 0)
 	resource->type = PRT_RESOURCE_TYPE_PROCSET;
     else if (prt_resfile_strncmp(offset, PRT_RESOURCE_ENCODING,
-				     (int)STRLEN(PRT_RESOURCE_ENCODING)) == 0)
+			     (int)STRLEN_LITERAL(PRT_RESOURCE_ENCODING)) == 0)
 	resource->type = PRT_RESOURCE_TYPE_ENCODING;
     else if (prt_resfile_strncmp(offset, PRT_RESOURCE_CMAP,
-					 (int)STRLEN(PRT_RESOURCE_CMAP)) == 0)
+				 (int)STRLEN_LITERAL(PRT_RESOURCE_CMAP)) == 0)
 	resource->type = PRT_RESOURCE_TYPE_CMAP;
     else
     {
@@ -1997,7 +1990,7 @@ prt_dsc_text(char *comment, char *text)
     prt_write_file(prt_line_buffer);
 }
 
-#define prt_dsc_atend(c)	prt_dsc_text((c), "atend")
+# define prt_dsc_atend(c)	prt_dsc_text((c), "atend")
 
     static void
 prt_dsc_ints(char *comment, int count, int *ints)
@@ -2347,7 +2340,7 @@ mch_print_init(
     struct prt_ps_encoding_S *p_mbenc_first;
     struct prt_ps_charset_S  *p_mbchar = NULL;
 
-#if 0
+# if 0
     /*
      * TODO:
      * If "forceit" is false: pop up a dialog to select:
@@ -2363,7 +2356,7 @@ mch_print_init(
      */
     if (forceit)
 	s_pd.Flags |= PD_RETURNDEFAULT;
-#endif
+# endif
 
     /*
      * Set up font and encoding.
@@ -2728,14 +2721,10 @@ mch_print_begin(prt_settings_T *psettings)
     struct prt_ps_resource_S *res_cmap;
     int		retval = FALSE;
 
-    res_prolog = (struct prt_ps_resource_S *)
-				      alloc(sizeof(struct prt_ps_resource_S));
-    res_encoding = (struct prt_ps_resource_S *)
-				      alloc(sizeof(struct prt_ps_resource_S));
-    res_cidfont = (struct prt_ps_resource_S *)
-				      alloc(sizeof(struct prt_ps_resource_S));
-    res_cmap = (struct prt_ps_resource_S *)
-				      alloc(sizeof(struct prt_ps_resource_S));
+    res_prolog = ALLOC_ONE(struct prt_ps_resource_S);
+    res_encoding = ALLOC_ONE(struct prt_ps_resource_S);
+    res_cidfont = ALLOC_ONE(struct prt_ps_resource_S);
+    res_cmap = ALLOC_ONE(struct prt_ps_resource_S);
     if (res_prolog == NULL || res_encoding == NULL
 	    || res_cidfont == NULL || res_cmap == NULL)
 	goto theend;
@@ -2753,7 +2742,7 @@ mch_print_begin(prt_settings_T *psettings)
 
     prt_dsc_textline("CreationDate", get_ctime(time(NULL), FALSE));
     prt_dsc_textline("DocumentData", "Clean8Bit");
-    prt_dsc_textline("Orientation", "Portrait");
+    prt_dsc_textline("Orientation", prt_portrait ? "Portrait" : "Landscape");
     prt_dsc_atend("Pages");
     prt_dsc_textline("PageOrder", "Ascend");
     // The bbox does not change with orientation - it is always in the default
@@ -2934,11 +2923,11 @@ mch_print_begin(prt_settings_T *psettings)
 	prt_dsc_resources(NULL, "encoding", buffer);
     }
     prt_dsc_requirements(prt_duplex, prt_tumble, prt_collate,
-#ifdef FEAT_SYN_HL
+# ifdef FEAT_SYN_HL
 					psettings->do_syntax
-#else
+# else
 					0
-#endif
+# endif
 					, prt_num_copies);
     prt_dsc_noarg("EndComments");
 
@@ -3119,7 +3108,7 @@ mch_print_end(prt_settings_T *psettings)
 
     // Write CTRL-D to close serial communication link if used.
     // NOTHING MUST BE WRITTEN AFTER THIS!
-    prt_write_file((char_u *)IF_EB("\004", "\067"));
+    prt_write_file((char_u *)"\004");
 
     if (!prt_file_error && psettings->outfile == NULL
 					&& !got_int && !psettings->user_abort)
@@ -3379,26 +3368,21 @@ mch_print_text_out(char_u *textp, int len UNUSED)
 	{
 	    // Convert non-printing characters to either their escape or octal
 	    // sequence, ensures PS sent over a serial line does not interfere
-	    // with the comms protocol.  Note: For EBCDIC we need to write out
-	    // the escape sequences as ASCII codes!
-	    // Note 2: Char codes < 32 are identical in EBCDIC and ASCII AFAIK!
-	    ga_append(&prt_ps_buffer, IF_EB('\\', 0134));
+	    // with the comms protocol.
+	    ga_append(&prt_ps_buffer, '\\');
 	    switch (ch)
 	    {
-		case BS:   ga_append(&prt_ps_buffer, IF_EB('b', 0142)); break;
-		case TAB:  ga_append(&prt_ps_buffer, IF_EB('t', 0164)); break;
-		case NL:   ga_append(&prt_ps_buffer, IF_EB('n', 0156)); break;
-		case FF:   ga_append(&prt_ps_buffer, IF_EB('f', 0146)); break;
-		case CAR:  ga_append(&prt_ps_buffer, IF_EB('r', 0162)); break;
-		case '(':  ga_append(&prt_ps_buffer, IF_EB('(', 0050)); break;
-		case ')':  ga_append(&prt_ps_buffer, IF_EB(')', 0051)); break;
-		case '\\': ga_append(&prt_ps_buffer, IF_EB('\\', 0134)); break;
+		case BS:   ga_append(&prt_ps_buffer, 'b'); break;
+		case TAB:  ga_append(&prt_ps_buffer, 't'); break;
+		case NL:   ga_append(&prt_ps_buffer, 'n'); break;
+		case FF:   ga_append(&prt_ps_buffer, 'f'); break;
+		case CAR:  ga_append(&prt_ps_buffer, 'r'); break;
+		case '(':  ga_append(&prt_ps_buffer, '('); break;
+		case ')':  ga_append(&prt_ps_buffer, ')'); break;
+		case '\\': ga_append(&prt_ps_buffer, '\\'); break;
 
 		default:
 			   sprintf((char *)ch_buff, "%03o", (unsigned int)ch);
-#ifdef EBCDIC
-			   ebcdic2ascii(ch_buff, 3);
-#endif
 			   ga_append(&prt_ps_buffer, ch_buff[0]);
 			   ga_append(&prt_ps_buffer, ch_buff[1]);
 			   ga_append(&prt_ps_buffer, ch_buff[2]);
@@ -3461,13 +3445,13 @@ mch_print_set_bg(long_u bgcol)
     void
 mch_print_set_fg(long_u fgcol)
 {
-    if (fgcol != (long_u)prt_fgcol)
-    {
-	prt_fgcol = (int)fgcol;
-	prt_attribute_change = TRUE;
-	prt_need_fgcol = TRUE;
-    }
+    if (fgcol == (long_u)prt_fgcol)
+	return;
+
+    prt_fgcol = (int)fgcol;
+    prt_attribute_change = TRUE;
+    prt_need_fgcol = TRUE;
 }
 
-# endif //FEAT_POSTSCRIPT
+#endif //FEAT_POSTSCRIPT
 #endif //FEAT_PRINTER
